#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sched.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>

#include "misc.h"

#include "mod.h"
#include "rule.h"
#include "branch.h"
#include "set.h"
#include "parse.h"

static void mod_pipe(STATE_SET * s, MOD_PIPE * md);
static MOD_PIPE *mod_new(STATE_SET * s, char *command, int timeout, int failmode);

static void mod_pipe(STATE_SET * s, MOD_PIPE * md)
{
  int inp[2], outp[2];
  int status;
  int inflags;
  int outflags;
  int efd;

  inp[0] = (-1);
  inp[1] = (-1);
  outp[0] = (-1);
  outp[1] = (-1);

  if (pipe(inp) || pipe(outp)) {
    parse_error(s, "error: unable to create pipe for %s: %s\n", md->m_command, strerror(errno));
    close(inp[0]);
    close(inp[1]);
    close(outp[0]);
    close(outp[1]);
    return;
  }
  md->m_pid = fork();
  switch (md->m_pid) {
  case -1:
    parse_error(s, "error: unable to fork %s: %s\n", md->m_command, strerror(errno));
    close(inp[0]);
    close(inp[1]);
    close(outp[0]);
    close(outp[1]);
    break;
  case 0:			/* in child */

#ifndef TRACE
    close(STDERR_FILENO);
    efd = open("/dev/null", O_WRONLY);
    if ((efd > 0) && (efd != STDERR_FILENO)) {
      dup2(efd, STDERR_FILENO);
      close(efd);
    }
#endif

    close(inp[1]);
    if (dup2(inp[0], STDIN_FILENO) != STDIN_FILENO) {	/* failure */
      /* WARNING: abuse of return code */
      exit(errno);
    }
    close(inp[0]);

    close(outp[0]);
    if (dup2(outp[1], STDOUT_FILENO) != STDOUT_FILENO) {	/* failure */
      /* WARNING: abuse of return code */
      exit(errno);
    }
    close(outp[1]);

    strexec(md->m_command);
    /* WARNING: abuse of return code */
    exit(errno);

    break;
  default:			/* in parent */
    close(inp[0]);
    close(outp[1]);

    md->m_outfd = inp[1];
    md->m_infd = outp[0];

    fcntl(md->m_infd, F_SETFD, FD_CLOEXEC);
    fcntl(md->m_outfd, F_SETFD, FD_CLOEXEC);

    inflags = fcntl(md->m_infd, F_GETFL, 0);
    outflags = fcntl(md->m_outfd, F_GETFL, 0);

    if ((inflags == (-1)) || (outflags == (-1)) || (fcntl(md->m_infd, F_SETFL, O_NONBLOCK | inflags) == (-1)) || (fcntl(md->m_outfd, F_SETFL, O_NONBLOCK | outflags) == (-1))) {
      parse_error(s, "error: unable to make pipe to %s nonblocking: %s\n", md->m_command, strerror(errno));
      close(md->m_infd);
      close(md->m_outfd);
    }
    /* yield to child */
    sched_yield();

    if (waitpid(md->m_pid, &status, WNOHANG)) {		/* found a child or error */
      if (WIFEXITED(status)) {
	parse_error(s, "error: unable to start child process %s: %s\n", md->m_command, strerror(WEXITSTATUS(status)));
      } else if (WIFSIGNALED(status)) {
	parse_error(s, "error: child process %s killed by signal %d\n", md->m_command, WTERMSIG(status));
      }
      close(md->m_infd);
      close(md->m_outfd);
    }
    break;
  }

}

static MOD_PIPE *mod_new(STATE_SET * s, char *command, int timeout, int failmode)
{
  MOD_PIPE *md;

  md = malloc(sizeof(MOD_PIPE));
  if (md) {
    md->m_failmode = failmode;
    md->m_times = timeout / 1000;
    md->m_timeu = (timeout * 1000) % 1000000;

    strncpy(md->m_command, command, IDSA_M_MEDIUM - 1);
    md->m_command[IDSA_M_MEDIUM - 1] = '\0';

#ifdef TRACE
    fprintf(stderr, "mod_new(): timemout is %ds + %us\n", md->m_times, md->m_timeu);
#endif

    md->m_interrupt = 0;
    md->m_count = 0;
    md->m_match = NULL;

    md->m_event = idsa_event_new(0);
    if (md->m_event) {
      /* if mod_pipe fails, return error via s->s_err* */
      mod_pipe(s, md);
    } else {
      free(md);
      md = NULL;
    }
  }
  return md;
}

void mod_free(MOD_PIPE * md)
{
  if (md->m_match) {
    free(md->m_match);
    md->m_match = NULL;
  }
  md->m_count = 0;

  if (md->m_event) {
    idsa_event_free(md->m_event);
    md->m_event = NULL;
  }
  close(md->m_infd);
  md->m_infd = (-1);

  close(md->m_outfd);
  md->m_outfd = (-1);

  /* a kill(md->m_pid) would be excessive */

  free(md);
}

int mod_trigger(MOD_PIPE * md, IDSA_EVENT * e)
{
  int result = 0;
  int fail = 1;
  int i;
  int sw, wr, rr, fr;
  char buffer[IDSA_M_MESSAGE];
  fd_set fs;
  struct timeval tv;

#ifdef TRACE
  fprintf(stderr, "mod_trigger(): triggering\n");
#endif

  if (md->m_interrupt) {	/* we got interrupted, read any junk, nonblocking */
    read(md->m_infd, buffer, IDSA_M_MESSAGE);
  }
  sw = idsa_event_tobuffer(e, buffer, IDSA_M_MESSAGE);
  if (sw > 0) {
#ifdef TRACE
    fprintf(stderr, "mod_trigger(): converted event to buffer of %d, writing to %d\n", sw, md->m_outfd);
#endif

    tv.tv_sec = md->m_times;
    tv.tv_usec = md->m_timeu;

    wr = write(md->m_outfd, buffer, sw);
    if (((wr < 0) && ((errno == EAGAIN) || (errno == EINTR))) || ((wr > 0) && (sw > wr))) {	/* try harder */
      FD_ZERO(&fs);
      FD_SET(md->m_outfd, &fs);
      if (select(md->m_outfd + 1, NULL, &fs, NULL, &tv) != 0) {
	if (wr < 0) {
	  wr = 0;
	}
	/* bit icky, since write could return -1 */
	wr += write(md->m_outfd, buffer + wr, sw - wr);
      }
    }
#ifdef TRACE
    fprintf(stderr, "mod_trigger(): should write %d, write result %d\n", sw, wr);
#endif

    if (wr == sw) {		/* write went ok */

      rr = read(md->m_infd, buffer, IDSA_M_MESSAGE);
      if ((rr < 0) && ((errno == EAGAIN) || (errno == EINTR))) {	/* try harder */
	FD_ZERO(&fs);
	FD_SET(md->m_infd, &fs);
	if (select(md->m_infd + 1, &fs, NULL, NULL, &tv) != 0) {
	  rr = read(md->m_infd, buffer, IDSA_M_MESSAGE);
	}
      }
#ifdef TRACE
      fprintf(stderr, "mod_trigger(): read result is %d\n", rr);
#endif

      if (rr > 0) {		/* something to read */
	fr = idsa_event_frombuffer(md->m_event, buffer, rr);
	if (fr == rr) {		/* exactly one event, not more, not less */
	  fail = 0;
	  if (md->m_event->e_result & IDSA_L_DENY) {
	    for (i = 0; i < md->m_count; i++) {
	      match_fire_false(md->m_match[i]);
	    }
	  } else {
	    for (i = 0; i < md->m_count; i++) {
	      match_fire_true(md->m_match[i]);
	    }
	  }
	}
      }
    }				/* else write failed */
  }				/* else event could not be converted to buffer */
  if (fail) {			/* some failure, abort */
    kill(md->m_pid, SIGINT);	/* tell child to wait for next event */
    md->m_interrupt = 1;

    if (md->m_failmode == MOD_FAILOPEN) {
      for (i = 0; i < md->m_count; i++) {
	match_fire_true(md->m_match[i]);
      }
    } else {
      for (i = 0; i < md->m_count; i++) {
	match_fire_false(md->m_match[i]);
      }
    }
  } else {
    md->m_interrupt = 0;
  }

  return result;
}

RULE_MATCH *mod_match_make(STATE_SET * s, char *command, int timeout, int failmode)
{
  UNION_BRANCH *b;
  MOD_PIPE *md;
  RULE_MATCH *m, **q;

  m = NULL;
  md = NULL;
  b = s->s_branch;
  while (b) {
    if (b->b_type == BRANCH_TYPE_TABLE) {
      md = b->b_un.b_mod;
      if (strncmp(md->m_command, command, IDSA_M_MEDIUM)) {
	b = b->b_next;
	md = NULL;
      } else {
#ifdef TRACE
	fprintf(stderr, "mod_match_make(): match\n");
#endif
	b->b_last = s->s_currentrule;
	b = NULL;
      }
    } else {
      b = b->b_next;
    }
  }

  if (md == NULL) {		/* not found, create */
    md = mod_new(s, command, timeout, failmode);
    if (md) {
      b = branch_new(s->s_currentrule, BRANCH_TYPE_MOD);
      if (b) {
	b->b_un.b_mod = md;
	branch_append(s, b);
      } else {
	parse_errormem(s, sizeof(UNION_BRANCH));
	mod_free(md);
	md = NULL;
      }
    } else {
      parse_errormem(s, sizeof(MOD_PIPE));
    }
  }
  if (md) {			/* add match to this entry */
    m = match_new(RULE_GATE_NONE);
    if (m) {
#ifdef TRACE
      fprintf(stderr, "mod_match_make(): about to resize %p from %d to %d\n", md->m_match, md->m_count, md->m_count + 1);
#endif
      q = realloc(md->m_match, sizeof(RULE_MATCH *) * (md->m_count + 1));
      if (q) {
	q[md->m_count] = m;
	md->m_count++;
	md->m_match = q;
      } else {
	parse_errormem(s, sizeof(RULE_MATCH *) * (md->m_count + 1));
	match_free(m);
	m = NULL;
      }
    } else {
      parse_errormem(s, sizeof(RULE_MATCH));
    }
  }
  return m;
}
