#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include <idsa_internal.h>

#include "types.h"
#include "rule.h"
#include "table.h"
#include "branch.h"
#include "set.h"
#include "parse.h"

static TABLE_UNIT *table_insert(TABLE_HEAD * h, IDSA_UNIT * u);

/****************************************************************************/

void table_free(TABLE_HEAD * h)
{
  int i;
  TABLE_UNIT *u;
  TABLE_HEAD *t;

  while (h != NULL) {

    if (h->h_units != NULL) {
      for (i = 0; i < h->h_ucount; i++) {
	u = &(h->h_units[i]);
	if (u->u_unit != NULL) {
	  idsa_unit_free(u->u_unit);
	  u->u_unit = NULL;
	}
	if (u->u_match != NULL) {
	  free(u->u_match);
	  u->u_match = NULL;
	}
	u->u_count = 0;
      }
      free(h->h_units);
      h->h_units = NULL;
    }
    h->h_ucount = 0;

    t = h->h_intersect;
    free(h);
    h = t;
  }
}

RULE_MATCH *table_match_make(STATE_SET * s, unsigned int t, char *n, char *p)
{
  UNION_BRANCH *b;
  TABLE_HEAD *th;
  TABLE_UNIT *tu;
  IDSA_UNIT *u;
  RULE_MATCH *m, **q;

  b = s->s_branch;
  m = NULL;
  th = NULL;

  while (b) {
    if (b->b_type == BRANCH_TYPE_TABLE) {
      th = b->b_un.b_table;
      if ((strncmp(th->h_name, n, IDSA_M_NAME) == 0) && ((th->h_type == t) || (t == IDSA_T_NULL))) {
#ifdef TRACE
	t = th->h_type;
	fprintf(stderr, "table_make_match(): found matching name <%s:0x%04x>\n", n, t);
#endif
	b->b_last = s->s_currentrule;
	b = NULL;		/* bomb - a break in the wrong place blew up ariane */
      } else {
	b = b->b_next;
	th = NULL;
      }
    } else {
      b = b->b_next;
    }
  }

  if (th == NULL) {		/* not found, better create */
    if (t != IDSA_T_NULL) {
      th = table_new(t, n);
      if (th) {
	b = branch_new(s->s_currentrule, BRANCH_TYPE_TABLE);
	if (b) {
	  b->b_un.b_table = th;
	  branch_append(s, b);
	} else {
	  parse_errormem(s, sizeof(UNION_BRANCH));
	  table_free(th);
	  th = NULL;
	}
      } else {
	parse_errormem(s, sizeof(TABLE_HEAD));
      }
    } else {
      parse_error(s, "require a type for %s", n);
    }
  }
  if (th) {
    u = idsa_unit_new(n, t, p);
    if (u) {
      tu = table_insert(th, u);
      if (tu) {
	m = match_new(RULE_GATE_NONE);
	if (m) {
#ifdef TRACE
	  fprintf(stderr, "table_match_make(): about to resize %p from %d to %d\n", tu->u_match, tu->u_count, tu->u_count + 1);
#endif
	  q = realloc(tu->u_match, sizeof(RULE_MATCH *) * (tu->u_count + 1));
	  if (q) {
	    q[tu->u_count] = m;
	    tu->u_count++;
	    tu->u_match = q;
	  } else {
	    parse_errormem(s, sizeof(RULE_MATCH *) * (tu->u_count + 1));
	    match_free(m);
	    m = NULL;
	  }
	} else {
	  parse_errormem(s, sizeof(RULE_MATCH));
	}
      } else {
	/* table_insert deletes u on failure */
	parse_error(s, "unable to insert %s:%s=%s", n, idsa_unit_name(t), p);
      }
    } else {
      parse_error(s, "unable to scan or allocate %s:%s=%s", n, idsa_unit_name(t), p);
    }
  }				/* else some failure and error set */
  return m;
}

int table_trigger(TABLE_HEAD * h, IDSA_EVENT * e)
{
  TABLE_UNIT *t;
  IDSA_UNIT *u;
  int i, j, k, r;
  int result = 0;

  if (h->h_reserved < idsa_required_count()) {
    u = idsa_event_unitbynumber(e, h->h_reserved);
  } else {
    u = idsa_event_unitbyname(e, h->h_name);
  }

  if (u && (h->h_type == u->u_type)) {
#ifdef TRACE
    fprintf(stderr, "table_trigger(): type 0x%04x\n", h->h_type);
#endif
    while (h != NULL) {

      i = 0;
      k = 0;
      j = h->h_ucount;
      t = NULL;

      while (i < j) {		/* binary search */
	k = i + ((j - i) / 2);
	t = &(h->h_units[k]);

	result++;
	r = idsa_unit_compare(u, t->u_unit);
#ifdef TRACE
	fprintf(stderr, "table_lookup(): Comparison of %d in range %d,%d yields %d\n", k, i, j, r);
#endif
	switch (r) {
	case IDSA_COMPARE_LESS:
	  t = NULL;
	  j = k;
	  break;
	case IDSA_COMPARE_EQUAL:
	case IDSA_COMPARE_INTERSECT:
	  i = k;
	  j = k;
	  break;
	case IDSA_COMPARE_MORE:
	  t = NULL;
	  i = k + 1;
	  break;
	}
      }

      /* we know matching position, now get to trigger things */
#ifdef TRACE
      fprintf(stderr, "table_trigger(): finished firing %d elements with match at %d\n", h->h_ucount, k);
#endif

      i = 0;
      if (t) {			/* found a match */
	while (i < k) {		/* fail all previous ones */
	  t = &(h->h_units[i]);
	  for (j = 0; j < t->u_count; j++) {
	    match_fire_false(t->u_match[j]);
	  }
	  i++;
	}
#ifdef TRACE
	fprintf(stderr, "table_trigger(): match at position %d\n", i);
#endif
	t = &(h->h_units[i]);	/* our match */
	for (j = 0; j < t->u_count; j++) {
	  match_fire_true(t->u_match[j]);
	}
	i++;
      }
      while (i < h->h_ucount) {	/* fail all subsequent ones */
	t = &(h->h_units[i]);
	for (j = 0; j < t->u_count; j++) {
	  match_fire_false(t->u_match[j]);
	}
	i++;
      }
      h = h->h_intersect;
    }
  }
  return result;
}

int table_dump(TABLE_HEAD * h, FILE * f)
{
  int i, j;
  TABLE_UNIT *u;
  fprintf(f, "type 0x%04x, unit count %d\n", h->h_type, h->h_ucount);

  for (i = 0; i < h->h_ucount; i++) {
    u = &(h->h_units[i]);
    fprintf(f, "%d: type 0x%04x, ptr <", i, u->u_unit->u_type);
    for (j = 0; j < ((u->u_unit->u_type) & IDSA_S_MASK); j++) {
      if (isprint(u->u_unit->u_ptr[j])) {
	fputc(u->u_unit->u_ptr[j], f);
      } else {
	fprintf(f, "\\%02x", u->u_unit->u_ptr[j]);
      }
    }
    fprintf(f, ">, count %d\n", u->u_count);
  }

  return 0;
}

/****************************************************************************/

static TABLE_UNIT *table_insert(TABLE_HEAD * h, IDSA_UNIT * u)
{
  TABLE_UNIT *t;
  TABLE_HEAD *f, *g;
  int i, j, k, r;

  i = 0;
  f = h;
  g = h;
  r = IDSA_COMPARE_LESS;
  t = NULL;

#ifdef TRACE
  fprintf(stderr, "table_insert(): locating candidate position\n");
#endif

  while (f != NULL) {		/* go through h->h_intersect */
    i = 0;
    k = 0;
    j = f->h_ucount;

    r = IDSA_COMPARE_LESS;
    while (i < j) {		/* check in unit table */
      k = i + ((j - i) / 2);
      t = &(f->h_units[k]);

      r = idsa_unit_compare(u, t->u_unit);
#ifdef TRACE
      fprintf(stderr, "table_insert(): Comparison of %d in range %d,%d yields %d\n", k, i, j, r);
#endif
      switch (r) {
      case IDSA_COMPARE_LESS:
	j = k;
	break;
      case IDSA_COMPARE_INTERSECT:
      case IDSA_COMPARE_EQUAL:
	i = k;
	j = k;
	break;
      case IDSA_COMPARE_MORE:
	i = k + 1;
	break;
      }
    }
    switch (r) {
      /* these allow inserts in this table */
    case IDSA_COMPARE_LESS:
    case IDSA_COMPARE_EQUAL:
    case IDSA_COMPARE_MORE:
      f = NULL;
      break;
      /* need to go on to intersect */
    case IDSA_COMPARE_INTERSECT:
#ifdef TRACE
      fprintf(stderr, "table_insert(): intersection, need to continue with %p\n", f->h_intersect);
#endif
      f = f->h_intersect;
      g = f;
      i = 0;
      break;
    }
  }

  /* reached end, need to create new TABLE_HEAD */
  if (g == NULL) {
#ifdef TRACE
    fprintf(stderr, "table_insert(): Allocating new table head\n");
#endif
    g = table_new(h->h_type, h->h_name);
    if (g) {
      g->h_lookup = h->h_lookup;
      g->h_intersect = h->h_intersect;
      h->h_intersect = g;
    }
  }
  if (g) {
#ifdef TRACE
    fprintf(stderr, "table_insert(): will insert at position %d in table head %p\n", i, h);
#endif
    if (r == IDSA_COMPARE_EQUAL) {
      t = &(g->h_units[i]);
    } else {			/* not a complete match, need to make space */
      t = realloc(g->h_units, sizeof(TABLE_UNIT) * (g->h_ucount + 1));
      if (t) {
	g->h_units = t;
	if (i < g->h_ucount) {
#ifdef TRACE
	  fprintf(stderr, "table_insert(): moving %d to %d by one\n", i, g->h_ucount);
#endif
	  memmove(&g->h_units[i + 1], &g->h_units[i], sizeof(TABLE_UNIT) * (g->h_ucount - i));
	}
	g->h_ucount++;
	h->h_utotal++;

	t = &(g->h_units[i]);
	t->u_unit = u;
	t->u_match = NULL;
	t->u_count = 0;

	u = NULL;
      }
    }
  } else {
    t = NULL;
  }
  if (u) {			/* on failure delete unit */
    idsa_unit_free(u);
  }
  return t;
}

TABLE_HEAD *table_new(unsigned int t, char *n)
{
  TABLE_HEAD *h;

#ifdef TRACE
  fprintf(stderr, "table_new(): new table for type 0x%04x\n", t);
#endif

  h = malloc(sizeof(TABLE_HEAD));
  if (h) {
    h->h_lookup = idsa_unit_lookup(t);

    if (h->h_lookup) {
      h->h_type = t;
      h->h_reserved = idsa_reserved_numberbyname(n);

      strncpy(h->h_name, n, IDSA_M_NAME - 1);
      h->h_name[IDSA_M_NAME - 1] = '\0';

      h->h_utotal = 0;
      h->h_ucount = 0;
      h->h_units = NULL;

      h->h_intersect = NULL;
    } else {
      free(h);
      h = NULL;
    }
  }
  return h;
}

TABLE_HEAD *table_find(UNION_BRANCH * b, char *n)
{
  TABLE_HEAD *result;

  while (b && b->b_type != BRANCH_TYPE_TABLE && strcmp(n, b->b_un.b_table->h_name)) {
    b = b->b_next;
  }

  if (b) {
    result = b->b_un.b_table;
  } else {
    result = NULL;
  }

  return result;
}
